import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

/** This file contains all the video routines for the emulator (well, this is the theory). <BR>
    This will make easier a port to a different API.<BR>
    Some code is strongly inspired by Jamicom (Thanks Ninn!!).
    Some code is strongly inspired by Darcnes (Thanks Nyef!!).
    This code is GPL. :o)
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    @version 0.2005 06/06/00
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "The Gnu Public License." */
final public class Video extends NESPictureProcessingUnit{
    
    /** Color palette by Matthew Conte. */
    protected final int rgb[] = { 0xFF808080, 0xFF0000BB, 0xFF3700BF, 0xFF8400A6, 0xFFBB006A, 0xFFB7001E, 0xFFB30000, 0xFF912600, 0xFF7B2B00, 0xFF003E00, 0xFF00480D, 0xFF003C22, 0xFF002F66, 0xFF000000, 0xFF050505, 0xFF050505, 0xFFC8C8C8, 0xFF0059FF, 0xFF443CFF, 0xFFB733CC, 0xFFFF33AA, 0xFFFF375E, 0xFFFF371A, 0xFFD54B00, 0xFFC46200, 0xFF3C7B00, 0xFF1E8415, 0xFF009566, 0xFF0084C4, 0xFF111111, 0xFF090909, 0xFF090909, 0xFFFFFFFF, 0xFF0095FF, 0xFF6F84FF, 0xFFD56FFF, 0xFFFF77CC, 0xFFFF6F99, 0xFFFF7B59, 0xFFFF915F, 0xFFFFA233, 0xFFA6BF00, 0xFF51D96A, 0xFF4DD5AE, 0xFF00D9FF, 0xFF666666, 0xFF0D0D0D, 0x0D0D0D, 0xFFFFFFFF, 0xFF84BFFF, 0xFFBBBBFF, 0xFFD0BBFF, 0xFFFFBFEA, 0xFFFFBFCC, 0xFFFFC4B7, 0xFFFFCCAE, 0xFFFFD9A2, 0xFFCCE199, 0xFFAEEEB7, 0xFFAAF7EE, 0xFFB3EEEF, 0xFFDDDDDD, 0xFF111111, 0xFF111111 };
    
    /** Cache for graphic patterns. <BR>
	The cache represents the patterns stored in memory to save time when the console draws on screen.<BR>
	This will be certainly modified with mappers implementation later.<BR>
	Cache is [512][8][8], it has 2 banks of 256 elements which are 8x8 pixels tiles. */
    private final int PatternCache[][][];

    /** Palette cache. <BR>
	Cache for color palette holding RGB values. <BR>
	It has 0x20 entries corresponding to the color palette.*/
    protected final int PaletteCache[];

    /** Video buffer. <BR>
	Buffer that represents the whole image (256x240 pixels).*/
    private final int videoBuffer[];

    /** Canvas. <BR>
	Canvas attached to the frame to paint inside.*/
    protected final Canvas canvas;

    /** Graphics. <BR>
	Used to paint into the canvas.*/
    private final Graphics g;

    /** Image. <BR>
	Image corresponding to the video buffer, it is used by graphics to paint. */
    private final Image offScreenImage;

    /** Frame. <BR>
	Frame used to show image on screen in a window (frame). */
    private final Frame f;

    /** Line buffer. <BR>
	Buffer representing the current line (256 pixels). */
    private final int lineBuffer[];

    /** Constructor. */
    public Video()
    {
	PatternCache = new int[ 512 ][ 8 ][ 8 ]; // 2 banks of 8x8 pixels tiles.
	PaletteCache = new int[ 0x20 ]; // The 0x20 color palette entries.
	videoBuffer = new int[ 61440 ]; // The whole image 256 * 240 pixels.
	lineBuffer = new int[ 256 ]; // The line is 256 pixels wide.
	// The following code initialize the video part on screen.
	canvas = new Canvas();
	canvas.setSize( 256, 240 );
	f = new Frame( "Javanese" );
	f.add( canvas );
	f.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {System.exit(0);}
	    });
	f.pack();
	f.setSize( new Dimension( 266 , 275 ) );
	f.show();
	g = canvas.getGraphics();
	offScreenImage = canvas.createImage(new MemoryImageSource(256, 224, videoBuffer, 0, 256));
    }

    /** This method fill the pattern cache. */
    final protected void caching()
    {
	byte Color = 0;
	char Result1;
	char Result2;
	for ( int i = 0; i < 0x2000 ; i += 16 )
	    {
		for ( int j = 0 ; j < 8 ; j ++ )
		    {
			Result1 = Memory[ i + j ];
			Result2 = Memory[ i + j + 0x0008 ];
			for ( int k = 0; k < 8 ; k ++ )
			    {
				if ( ( ( Result1 >>> k ) & 0x0001 ) == 0x0001 )
				    {
					Color ++;
				    }
				if ( ( ( Result2 >>> k ) & 0x0001 ) == 0x0001 )
				    {
					Color += 2;
				    }
				PatternCache[ i / 16 ][ j ][ 7 - k ] = Color;
				Color = 0;
			    }
		    }
	    }
    }
    
    /** This method renders the sprites on the current scanline. */
    protected final void renderSprites(){
	final int position[] = new int[ 64 ];
	final int colorIndex[] = new int[ 64 ];
	int counter = 0;
	int spriteOnScanline = 0;
	int sy;
	int TileIndex;
	int Attributes;
	int XCoordinate;
	int XPosition;
	boolean VerticalFlip;
	boolean HorizontalFlip;
	boolean BackgroundPriority;
	int pAddr;
	int pltNum;
	
	sy = Scanline - ( SpriteRam[ 0 ] + 1 );
	if( sy >= 0 && sy < SpriteSize )
	    {
		spriteOnScanline ++;
		VerticalFlip = ( SpriteRam[ 2 ] & 0x80 ) != 0;
		HorizontalFlip = ( SpriteRam[ 2 ] & 0x40 ) != 0;
		BackgroundPriority = ( SpriteRam[ 2 ] & 0x20 ) != 0;
		sy ^= !VerticalFlip ? 0 : 0x7;
		if ( SpriteSize == 8 )
		    {
			pAddr = ( char ) ( SpriteRam[ 1 ] + SpritePatternTableAddress );
		    }
		else
		    {
			pAddr = ( char ) ( ( ( SpriteRam[ 1 ] & 1 ) << 8 ) + ( SpriteRam[ 1 ] & ~1 ) );
			pAddr ^= ( sy < 8 ) ? 0 : 0x10;
			pAddr ^= !VerticalFlip ? 0 : 0x10;
			sy &= 0x7;
		    }
		for (int sx = 0; sx < 8; sx++)
		    {
			XPosition = SpriteRam[ 3 ] + sx;
			if ( XPosition >= 256 )
			    {
				break;
			    }
			if ( ( XPosition < SpriteClipping ) || ( BackgroundPriority && ( ( lineBuffer[ XPosition ] & 0x3 ) != 0 ) ) )
			    {
				continue;
			    }
			pltNum = PatternCache[ pAddr ][ sy ][ !HorizontalFlip ? sx : 7 - sx ];
			if( pltNum != 0 )
			    {
				pltNum |= ( SpriteRam[ 2 ] & 0x03 ) << 2;
				position[ counter ]= XPosition;
				colorIndex[ counter ] = pltNum + 0x10;
				counter ++;
				WasSpriteZeroHit = true;
			    }
		    }
	    }
	
	for( int i = 4; i < 256; i += 4 )
	    {
		sy = Scanline - ( SpriteRam[ i ] + 1 );
		if( sy >= 0 && sy < SpriteSize )
		    {
			if ( spriteOnScanline < 8 )
			    {
				spriteOnScanline ++;
			    }
			else
			    {
				IsThereMoreThanEightSpritesOnScanline = true;
				break;
			    }
			TileIndex = SpriteRam[ i + 1 ];
			Attributes = SpriteRam[ i + 2 ];
			XCoordinate = SpriteRam[ i + 3 ];
			VerticalFlip = ( Attributes & 0x80 ) != 0;
			HorizontalFlip = ( Attributes & 0x40 ) != 0;
			BackgroundPriority = ( Attributes & 0x20 ) != 0;
			sy ^= !VerticalFlip ? 0 : 0x7;
			if ( SpriteSize == 8 )
			    {
				pAddr = ( char ) ( TileIndex + SpritePatternTableAddress );
			    }
			else
			    {
				pAddr = ( char ) ( ( ( TileIndex & 1 ) << 8 ) + ( TileIndex & ~1 ) );
				pAddr ^= ( sy < 8 ) ? 0 : 0x10;
				pAddr ^= !VerticalFlip ? 0 : 0x10;
				sy &= 0x7;
			    }
			for (int sx = 0; sx < 8; sx++)
			    {
				XPosition = XCoordinate + sx;
				if ( XPosition >= 256 )
				    {
					break;
				    }
				if ( ( XPosition < SpriteClipping ) || ( BackgroundPriority && ( ( lineBuffer[ XPosition ] & 0x3 ) != 0 ) ) )
				    {
					continue;
				    }
				pltNum = PatternCache[ pAddr ][ sy ][ !HorizontalFlip ? sx : 7 - sx ];
				if( pltNum != 0 )
				    {
					pltNum |= ( Attributes & 0x03 ) << 2;
					position[ counter ]= XPosition;
					colorIndex[ counter ] = pltNum + 0x10;
					counter ++;
				    }
			    }
		    }
	    }

	for( int i = counter - 1; i >= 0; i --)
	    {
		lineBuffer[ position[ i ] ] = colorIndex[ i ];
	    }
    }
    
    /** This method renders the background. */
    protected final void renderBackground(){
	int paletteIndex;
	int XScroll = VRAMAddressRegister & 0x1F;
	int YOffset = ( VRAMAddressRegister >>> 12 ) & 7;
	int AddressComplement = VRAMAddressRegister & 0x3FF; // YScrollXscroll
	int Address = 0x2000;
	int NameTableAddressNumber = ( VRAMAddressRegister >>> 10 ) & 3;
	int indexJ;
	int index;
	int attributeBase = ( ( VRAMAddressRegister >>> 4 ) & 4 ) | ( XScroll & 2 );
	int attributeLine;
	switch( NameTableAddressNumber )
	    {
	    case 0 : attributeLine = 0x23C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 ) + ( XScroll >>> 2 );
		break;
	    case 1 : attributeLine = NameTable1 + 0x3C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 ) + ( XScroll >>> 2 );
		break;
	    case 2 : attributeLine = NameTable2 + 0x3C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 ) + ( XScroll >>> 2 );
		break;
	    default : attributeLine = 0x27C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 ) + ( XScroll >>> 2 );
	    }
	for( int j = 0; j < 33; j ++ )
	    {
		switch( NameTableAddressNumber )
		    {
		    case 0 : Address = 0x2000 | AddressComplement; //a revoir pour les autres mappers
			break;
		    case 1 : Address = NameTable1 | AddressComplement;
			break;
		    case 2 : Address = NameTable2 | AddressComplement;
			break;
		    default : Address = 0x2400 | AddressComplement; //a revoir pour les autres mappers
		    }
		indexJ = j<<3;
		for( int i = 0; i < 8 ; i ++ )
		    {
			paletteIndex = PatternCache[ Memory[ Address ] + BackgroundPatternTableAddress ][ YOffset ][ i ];
			paletteIndex |= ( ( Memory[ attributeLine ] >>> attributeBase ) & 3 )<< 2;
			index = indexJ + i - XOffset;
			if ( index >= BackgroundClipping && index < 256 )
			    {
				lineBuffer[ index ] = paletteIndex;//opti
			    }
		    }
		if( ( XScroll & 1 ) != 0 ) attributeBase ^= 2;
		if( ( XScroll & 3 ) == 3 ) attributeLine ++;
		XScroll ++;
		AddressComplement ++;
		if( XScroll > 31 )
		    {
			XScroll = 0;
			NameTableAddressNumber = ( ( VRAMAddressRegister >>> 10 ) & 3)^ 1;
			AddressComplement = VRAMAddressRegister & 0x3E0;
			switch( NameTableAddressNumber )
			    {
			    case 0 : attributeLine = 0x23C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 );
				break;
			    case 1 : attributeLine = NameTable1 + 0x3C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 );
				break;
			    case 2 : attributeLine = NameTable2  + 0x3C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 );
				break;
			    default : attributeLine = 0x27C0 + ( ( VRAMAddressRegister >>> 4 ) & 0x38 );
			    }
		    }
		//System.out.println("j:XS "+j+":"+XScroll);
	    }
	//System.out.println("ligne: "+(int)Scanline);
	//System.out.println("XScroll: "+XScroll+" Rd: "+(int)RefreshData+" Yoff: "+YOffset);
    }
    
    /** Renders image on screen. */
    protected final void drawDisplay() {
	offScreenImage.flush();
	g.drawImage(offScreenImage, 0, 0, null);
    }

    /** Renders scanline in video buffer. */
    protected final void drawLine(){
	if( BackgroundVisibility )
	    {
		renderBackground();
	    }
	{
	    //if( Scanline == SpriteRam[0]+ 7 ) WasSpriteZeroHit = true;
	}
	if( SpriteVisibility )
	    {
		renderSprites();
	    }
	final int scanlineIndex = ( Scanline - 8  ) << 8; // in order to optimize //there is a strange one tile scroll here
	if( scanlineIndex >= 0 && scanlineIndex  <= 0xF000)
	    {
		for( int i = 0; i < 256; i++)
		    {
			videoBuffer[ scanlineIndex + i ] = PaletteCache[ lineBuffer[ i ] ];
		    }
	    }
	java.util.Arrays.fill( lineBuffer, 0 ); // Empty line buffer.
	VRAMAddressRegister += 0x1000;
	if( ( VRAMAddressRegister & 0x8000 ) != 0 )
	    {
		final int tempo = ( VRAMAddressRegister & 0x03E0 ) + 32;
		VRAMAddressRegister &= 0x7C1F;
		if( tempo == 0x03C0 )
		    {
			VRAMAddressRegister ^= 0x800;
		    }
		else
		    {
			VRAMAddressRegister |= ( tempo & 0x03E0 );
		    }
	    }
    }    
}
